#include "dynamic_internal_representation.h"

#include "module.h"
#include "internal_representation_helper.h"
#include "../tools/misc.h"

extern DynamicModule *ast_root;

void DynamicEventList::class_name()
{
  std::cout << "DynamicEventList" << std::endl;
}

void DynamicPort::class_name()
{
  std::cout << "Port" << std::endl;
}

void DynamicPort::print(unsigned int indent)
{
  for(unsigned int i=0; i<indent; i++ ) {
    std::cout << "  ";
  }
  std::cout << name_ << ":" << address_ << std::endl;
}

void DynamicPointer::class_name()
{
  std::cout << "Pointer" << std::endl;
}

void DynamicPointer::print(unsigned int indent)
{
  for(unsigned int i=0; i<indent; i++ ) {
    std::cout << "  ";
  }
  std::cout << name_ << "->" << address_ << std::endl;
}

void DynamicReference::class_name()
{
  std::cout << "Reference" << std::endl;
}

void DynamicReference::print(unsigned int indent)
{
  for(unsigned int i=0; i<indent; i++ ) {
    std::cout << "  ";
  }
  std::cout << name_ << "@" << address_ << std::endl;
}

void DynamicVariable::class_name()
{
  std::cout << "Variable" << std::endl;
}

void DynamicVariable::print(unsigned int indent)
{
  for(unsigned int i=0; i<indent; i++ ) {
    std::cout << "  ";
  }
  std::cout << name_ << "$" << address_ << std::endl;
}

void DynamicEvent::class_name()
{
  std::cout << "Event" << std::endl;
}

void DynamicEvent::print(unsigned int indent)
{
  for(unsigned int i=0; i<indent; i++ ) {
    std::cout << "  ";
  }
  std::cout << name_ << "$" << address_ << std::endl;
}

void DynamicPrimChannel::class_name()
{
  std::cout << "DynamicPrimChannel" << std::endl;
}

void DynamicPrimChannel::print(unsigned int indent)
{
  for(unsigned int i=0; i<indent; i++ ) {
    std::cout << "  ";
  }
  std::cout << name_ << "%" << address_ << "&" << class_name_ << std::endl;
}

void DynamicEventOrListElement::class_name()
{
  std::cout << "DynamicEventOrAndListElement" << std::endl;
}

void DynamicEventOrListElement::print(unsigned int indent)
{
  for(unsigned int i=0; i<indent; i++ ) {
    std::cout << "  ";
  }
  std::cout << name_ << "<" << address_ << "{" << event_ << std::endl;
}

void DynamicEventOrList::class_name()
{
  std::cout << "DynamicEventOrList" << std::endl;
}

void DynamicEventOrList::print(unsigned int indent)
{
  for(unsigned int i=0; i<indent; i++ ) {
    std::cout << "  ";
  }
  std::cout << name_ << "<" << address_ << ":";

  for(std::list<int*>::iterator
      event_iter  = events_.begin();
      event_iter != events_.end();
      event_iter++) {

    std::cout << (*event_iter) << ", ";
  }
  std::cout << std::endl;
}

void DynamicEventAndListElement::class_name()
{
  std::cout << "DynamicEventAndListElement" << std::endl;
}

void DynamicEventAndListElement::print(unsigned int indent)
{
  for(unsigned int i=0; i<indent; i++ ) {
    std::cout << "  ";
  }
  std::cout << name_ << "<" << address_ << "{" << event_ << std::endl;
}

void DynamicEventAndList::class_name()
{
  std::cout << "DynamicEventAndList" << std::endl;
}

void DynamicEventAndList::print(unsigned int indent)
{
  for(unsigned int i=0; i<indent; i++ ) {
    std::cout << "  ";
  }
  std::cout << name_ << "<" << address_ << ":";

  for(std::list<int*>::iterator
      event_iter  = events_.begin();
      event_iter != events_.end();
      event_iter++) {

    std::cout << (*event_iter) << ", ";
  }
  std::cout << std::endl;
}

void DynamicModule::class_name()
{
  std::cout << "Module" << std::endl;
}

void DynamicModule::print_tree(unsigned int indent)
{
  for(unsigned int i=0; i<indent; i++ ) {
    std::cout << "  ";
  }
  std::cout << name_ << "#" << address_ << std::endl;

  for(std::list<DynamicModule*>::iterator
      iter  = sub_modules_.begin();
      iter != sub_modules_.end();
      iter++) {

    (*iter)->print_tree(indent+1);
  }

  for(std::list<DynamicHierarchicalChannel*>::iterator
      iter  = sub_hierarchical_channels_.begin();
      iter != sub_hierarchical_channels_.end();
      iter++) {

    (*iter)->print_tree(indent+1);
  }

  #define PRINT_OBJECT(TYPE, VARIABLE)    \
  for(std::list< TYPE *>::iterator        \
      iter_variable  = VARIABLE .begin(); \
      iter_variable != VARIABLE .end();   \
      iter_variable++) {                  \
    (*iter_variable)->print(indent+1);    \
  }

  PRINT_OBJECT(DynamicPort, ports_)
  PRINT_OBJECT(DynamicPointer, pointers_)
  PRINT_OBJECT(DynamicReference, references_)
  PRINT_OBJECT(DynamicVariable, variables_)
  PRINT_OBJECT(DynamicPrimChannel, sub_channels_)
  PRINT_OBJECT(DynamicEvent, events_)
  PRINT_OBJECT(DynamicEventAndList, event_and_lists_)
  PRINT_OBJECT(DynamicEventOrList, event_or_lists_)

  #undef PRINT_OBJECT
}

void DynamicHierarchicalChannel::class_name()
{
  std::cout << "HierarchicalChannel" << std::endl;
}

void DynamicHierarchicalChannel::print_tree(unsigned int indent)
{
  for(unsigned int i=0; i<indent; i++ ) {
    std::cout << "  ";
  }
  std::cout << name_ << "?" << address_ << std::endl;

  for(std::list<DynamicModule*>::iterator
      iter  = sub_modules_.begin();
      iter != sub_modules_.end();
      iter++) {

    (*iter)->print_tree(indent+1);
  }

  for(std::list<DynamicHierarchicalChannel*>::iterator
      iter  = sub_hierarchical_channels_.begin();
      iter != sub_hierarchical_channels_.end();
      iter++) {

    (*iter)->print_tree(indent+1);
  }

  #define PRINT_OBJECT(TYPE, VARIABLE)    \
  for(std::list< TYPE *>::iterator        \
      iter_variable  = VARIABLE .begin(); \
      iter_variable != VARIABLE .end();   \
      iter_variable++) {                  \
    (*iter_variable)->print(indent+1);    \
  }

  PRINT_OBJECT(DynamicPort, ports_)
  PRINT_OBJECT(DynamicPointer, pointers_)
  PRINT_OBJECT(DynamicReference, references_)
  PRINT_OBJECT(DynamicVariable, variables_)
  PRINT_OBJECT(DynamicPrimChannel, sub_channels_)
  PRINT_OBJECT(DynamicEvent, events_)
  PRINT_OBJECT(DynamicEventAndList, event_and_lists_)
  PRINT_OBJECT(DynamicEventOrList, event_or_lists_)

  #undef PRINT_OBJECT
}

DynamicObject*
get_dynamic_object_of_variable(
    DynamicModule *root,
    std::string host_module_type_name,
    int host_module_instance_id,
    std::string reference_name,
    int &instance_counter)
{
  // Here we are looking in the current modle. First, we have to check the
  // type name of the module
  if(root->name_ == host_module_type_name) {

    // Then, we check the instance id
    if(instance_counter == host_module_instance_id) {

      #define COMPARE_NAME(TYPE, VARIABLE)               \
      for(std::list< TYPE *>::iterator                   \
          iter_variable  = root-> VARIABLE .begin();     \
          iter_variable != root-> VARIABLE .end();       \
          iter_variable++) {                             \
        if((*iter_variable)->name_  == reference_name) { \
          return *iter_variable;                         \
        }                                                \
      }

      COMPARE_NAME(DynamicModule, sub_modules_)
      COMPARE_NAME(DynamicPort, ports_)
      COMPARE_NAME(DynamicPointer, pointers_)
      COMPARE_NAME(DynamicReference, references_)
      COMPARE_NAME(DynamicVariable, variables_)
      COMPARE_NAME(DynamicPrimChannel, sub_channels_)
      COMPARE_NAME(DynamicEvent, events_)
      COMPARE_NAME(DynamicEventAndList, event_and_lists_)
      COMPARE_NAME(DynamicEventOrList, event_or_lists_)

      #undef COMPARE_NAME
    }

    // In the case that the type was correct, however, the instance id was not
    // machtechint, we have to increase the instance counter
    instance_counter++;
  }

  // traversing over all submodules
  for(std::list<DynamicModule*>::iterator
      sub_module_iter  = root->sub_modules_.begin();
      sub_module_iter != root->sub_modules_.end();
      sub_module_iter++) {


    DynamicObject *mapped_variable
       = get_dynamic_object_of_variable(
           *sub_module_iter,
           host_module_type_name,
           host_module_instance_id,
           reference_name,
           instance_counter);

    // If the return value is not null, than we can stop the travesing and
    // return the address
    if(mapped_variable) {
      return mapped_variable;
    }
  }

  // traversing over all submodules
  for(std::list<DynamicHierarchicalChannel*>::iterator
      sub_channel_iter  = root->sub_hierarchical_channels_.begin();
      sub_channel_iter != root->sub_hierarchical_channels_.end();
      sub_channel_iter++) {


    DynamicObject *mapped_variable
       = get_dynamic_object_of_variable(
           *sub_channel_iter,
           host_module_type_name,
           host_module_instance_id,
           reference_name,
           instance_counter);

    // If the return value is not null, than we can stop the travesing and
    // return the address
    if(mapped_variable) {
      return mapped_variable;
    }
  }

  // In the case there was no matchting in the current (root) module and
  // all the submodules we have to return NULL
  return NULL;
}

bool
get_instance_id_of_hosting_module(
    DynamicModule *root,
    int *address_of_variable,
    std::string module_name,
    int &instance_counter)
{
  // Here we are checking the current module type of root and compare the type
  // name with the name of the host module
  if(module_name == root->name_) {

    // We are comparing all the variable addresses the
    // given address. If it is a matching, we can return.
    #define SEARACH_FOR_VARIABLE(TYPE, VARIABLE)               \
    for(std::list< TYPE *>::iterator                           \
        iter_variable  = root-> VARIABLE .begin();             \
        iter_variable != root-> VARIABLE .end();               \
        iter_variable++) {                                     \
      if((*iter_variable)->address_ == address_of_variable) {  \
        return true;                                           \
      }                                                        \
    }

    SEARACH_FOR_VARIABLE(DynamicModule, sub_modules_)
    SEARACH_FOR_VARIABLE(DynamicPort, ports_)
    SEARACH_FOR_VARIABLE(DynamicPointer, pointers_)
    SEARACH_FOR_VARIABLE(DynamicReference, references_)
    SEARACH_FOR_VARIABLE(DynamicVariable, variables_)
    SEARACH_FOR_VARIABLE(DynamicPrimChannel, sub_channels_)
    SEARACH_FOR_VARIABLE(DynamicEvent, events_)
    SEARACH_FOR_VARIABLE(DynamicEventAndList, event_and_lists_)
    SEARACH_FOR_VARIABLE(DynamicEventOrList, event_or_lists_)

    #undef SEARACH_FOR_VARIABLE

    // If the don't have a matching, we just have to increase the counter
    instance_counter++;
  }

  // Here we are traversing all the sub modules
  for(std::list<DynamicModule*>::iterator
      iter_sub_module  = root->sub_modules_.begin();
      iter_sub_module != root->sub_modules_.end();
      iter_sub_module++) {

    if(get_instance_id_of_hosting_module(*iter_sub_module, address_of_variable,
           module_name, instance_counter)) {

      // in this case we found the variable, so we can return
      return true;
    }
  }

  // Here we are traversing all the sub modules
  for(std::list<DynamicHierarchicalChannel*>::iterator
      iter_sub_channel  = root->sub_hierarchical_channels_.begin();
      iter_sub_channel != root->sub_hierarchical_channels_.end();
      iter_sub_channel++) {

    if(get_instance_id_of_hosting_module(*iter_sub_channel, address_of_variable,
           module_name, instance_counter)) {

      // in this case we found the variable, so we can return
      return true;
    }
  }

  return false;
}

std::pair<DynamicModule*, std::string>
get_host_module_and_name_of_variable(DynamicModule *root, int *variable)
{
  // here we are iterating over all of the variables of the current module
  // root. If the address is matching, we can return the current module
  // root and the name of the variable

  #define COMPARE_NAME(TYPE, VARIABLE)               \
  for(std::list< TYPE *>::iterator                   \
      iter_variable  = root-> VARIABLE .begin();     \
      iter_variable != root-> VARIABLE .end();       \
      iter_variable++) {                             \
    if((*iter_variable)->address_  == variable) {    \
      return std::pair<DynamicModule*, std::string>( \
                 root, (*iter_variable)->name_);     \
    }                                                \
  }

  COMPARE_NAME(DynamicModule, sub_modules_);
  COMPARE_NAME(DynamicPort, ports_);
  COMPARE_NAME(DynamicPointer, pointers_);
  // address is the address of the mapped reference.
  // COMPARE_NAME(DynamicReference, references_);
  COMPARE_NAME(DynamicVariable, variables_);
  COMPARE_NAME(DynamicPrimChannel, sub_channels_);
  COMPARE_NAME(DynamicEvent, events_);
  COMPARE_NAME(DynamicEventAndList, event_and_lists_);
  COMPARE_NAME(DynamicEventOrList, event_or_lists_);

  #undef COMPARE_NAME

  // Traversing the module hierarchy
  for(std::list<DynamicModule*>::iterator
      sub_module_iter  = root->sub_modules_.begin();
      sub_module_iter != root->sub_modules_.end();
      sub_module_iter++) {

    std::pair<DynamicModule*, std::string> result
      = get_host_module_and_name_of_variable(*sub_module_iter, variable);

    if(result.first) {
      return result;
    }
  }

  // Traversing the module hierarchy
  for(std::list<DynamicHierarchicalChannel*>::iterator
      sub_channel_iter  = root->sub_hierarchical_channels_.begin();
      sub_channel_iter != root->sub_hierarchical_channels_.end();
      sub_channel_iter++) {

    std::pair<DynamicModule*, std::string> result
      = get_host_module_and_name_of_variable(*sub_channel_iter, variable);

    if(result.first) {
      return result;
    }
  }

  return std::pair<DynamicModule*, std::string> (NULL, "");
}

void
get_all_components_of_type(
    DynamicModule *root,
    std::string type_name,
    std::list<DynamicModule*> &modules)
{
  if(root == NULL) {
    return;
  }

  if(root->name_ == type_name) {
    modules.push_back(root);
  }

  // Traversing the module hierarchy
  for(std::list<DynamicModule*>::iterator
      sub_module_iter  = root->sub_modules_.begin();
      sub_module_iter != root->sub_modules_.end();
      sub_module_iter++) {

    get_all_components_of_type(*sub_module_iter, type_name, modules);
  }

  // Traversing the module hierarchy
  for(std::list<DynamicHierarchicalChannel*>::iterator
      sub_channel_iter  = root->sub_hierarchical_channels_.begin();
      sub_channel_iter != root->sub_hierarchical_channels_.end();
      sub_channel_iter++) {

    get_all_components_of_type(*sub_channel_iter, type_name, modules);
  }
}

DynamicModule*
get_host_module_of_member(DynamicModule *root, int *address_of_variable)
{
  // Traversing the module hierarchy
  for(std::list<DynamicPrimChannel*>::iterator
      sub_channel_iter  = root->sub_channels_.begin();
      sub_channel_iter != root->sub_channels_.end();
      sub_channel_iter++) {

    if((*sub_channel_iter)->address_ == address_of_variable) {
      return root;
    }
  }

  if(root->address_ == address_of_variable)
    return root;
//  #define SERACH_FOR_ADDRESS(TYPE, VARIABLE)               \
//  for(std::list< TYPE *>::iterator                           \
//      iter_variable  = root-> VARIABLE .begin();             \
//      iter_variable != root-> VARIABLE .end();               \
//      iter_variable++) {                                     \
//    if((*iter_variable)->address_ == address_of_variable) {  \
//      return root;                                           \
//    }                                                        \
//  }
//
//  SEARACH_FOR_VARIABLE(DynamicPort, ports_)
//  SEARACH_FOR_VARIABLE(DynamicPointer, pointers_)
//  SEARACH_FOR_VARIABLE(DynamicVariable, variables_)
//  SEARACH_FOR_VARIABLE(DynamicEvent, events_)
//  SEARACH_FOR_VARIABLE(DynamicEventAndList, event_and_lists_)
//  SEARACH_FOR_VARIABLE(DynamicEventOrList, event_or_lists_)
//
//  #undef SERACH_FOR_ADDRESS

  // Traversing the module hierarchy
  for(std::list<DynamicModule*>::iterator
      sub_module_iter  = root->sub_modules_.begin();
      sub_module_iter != root->sub_modules_.end();
      sub_module_iter++) {

    DynamicModule *module
      = get_host_module_of_member(*sub_module_iter, address_of_variable);

    if(module) {
      return module;
    }
  }

  // Traversing the module hierarchy
  for(std::list<DynamicHierarchicalChannel*>::iterator
      sub_channel_iter  = root->sub_hierarchical_channels_.begin();
      sub_channel_iter != root->sub_hierarchical_channels_.end();
      sub_channel_iter++) {

    DynamicModule *module
      = get_host_module_of_member(*sub_channel_iter, address_of_variable);

    if(module) {
      return module;
    }
  }

  return NULL;
}

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
