#ifndef PORT_MAPPING_H_INCLUDED_
#define PORT_MAPPING_H_INCLUDED_

#include <list>

#include "rose.h"

#include "instance_tree.h"

namespace risc {

class Port;
class Module;

/**
 * \fn       get_mapped_channel_instances
 * \brief    This function determines to which channels this port can be bound.
 *           An instance of the given port can occur multiple times.
 */
std::list<InstanceTree> get_mapped_channel_instances(Port *port);

/**
 * \fn       get_mapped_channel
 * \brief    This function determines to which channel this port is bound
 */
InstanceTree get_mapped_channel_instance(InstanceTree port);

/**
 * \fn       get_mapped_instance
 * \brief    This function is only for internal use
 *           Here we determine to which object (port or channel) the given
 *           port is mapped in the parent module
 */
SgVariableSymbol* get_mapped_symbol(
    SgVariableSymbol *port,
    Module *parent_module,
    SgVariableDefinition *parent_variable);

} // end of namespace risc

#endif /* PORT_MAPPING_H_INCLUDED_ */

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
