#ifndef PRIMITIVE_CHANNEL_INSTANCE_H_INCLUDED_
#define PRIMITIVE_CHANNEL_INSTANCE_H_INCLUDED_

#include <vector>

#include "rose.h"

#include "class.h" // included for get_definition
#include "primitive_channel.h" // included for get_definition

#include "instance.h"

namespace risc {

class PrimitiveChannelInstance: public Instance {

public:

  /**
   * \brief Default constructor
   */
  explicit PrimitiveChannelInstance(SgVariableDefinition *ast_node,
      PrimitiveChannel *definition);

  /**
   * \brief Get function for the name of the ast type of the channel instance
   * \return Name as std::string
   */
  virtual std::string get_ast_type_name();


  /**
   * \fn       get_definition
   * \brief    returns this associated definition of the primitive channel
   *           instance
   */
  virtual PrimitiveChannel* get_definition();

private:

  /**
   * \brief This pointer points to the corresponding definition in the
   *        internal representation
   */
  PrimitiveChannel *definition_;

  /**
   * \brief Disabled copy constructor
   */
  PrimitiveChannelInstance(const PrimitiveChannelInstance &pci);
};

typedef std::vector<PrimitiveChannelInstance*>
            PrimitiveChannelInstanceVector;
typedef std::vector<PrimitiveChannelInstance*>::iterator
            PrimitiveChannelInstanceVectorIter;
typedef std::vector<PrimitiveChannelInstance*>::const_iterator
            PrimitiveChannelInstanceVectorConstIter;
typedef std::vector<PrimitiveChannelInstance*>::iterator
            PrimitiveChannelInstanceVectorIter;

} // end namespace risc

#endif /* PRIMITIVE_CHANNEL_INSTANCE_H_INCLUDED_ */

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
