%{

/*
 * Parser.y file
 * To generate the parser run: "bison Parser.y"
 */

%}

%code requires {

#ifndef FUNCTION_ANNOTATION_PARSER_TYPEDEF_FUNCTION_ANNOTATION_PARSER_SCANNER_T
#define FUNCTION_ANNOTATION_PARSER_TYPEDEF_FUNCTION_ANNOTATION_PARSER_SCANNER_T

#include <stdio.h>
#include <iostream>
#include <cstdlib>
#include "function_annotation.h"


extern FunctionAnnotation *function_annotation;

void function_annotation_parsererror(const char *s);
int function_annotation_parserlex();

#ifdef __cplusplus

/* The "const" storage-class-modifier is valid. */
#define YY_USE_CONST

#else	/* ! __cplusplus */

/* C99 requires __STDC__ to be defined as 1. */
#if defined (__STDC__)

#define YY_USE_CONST

#endif	/* defined (__STDC__) */
#endif	/* ! __cplusplus */

#ifdef YY_USE_CONST
#define yyconst const
#else
#define yyconst
#endif

typedef struct yy_buffer_state *YY_BUFFER_STATE;
YY_BUFFER_STATE function_annotation_parser_scan_string (yyconst char *yy_str  );

#endif
}

%union {
  long long number_;
  bool is_conflict_free_;
  char *function_name_;
  WaitAnnotation *wait_annotation_;
  AnnotatedTime *annotated_time_;
  WAIT_CONSTRUCT wait_;
  FunctionAnnotation *function_annotation_;
}

/* declaration of tokens */
%token                        T_LPAREN
%token                        T_RPAREN
%token                        T_COMMA
%token                        T_RISC
%token                        T_TIME

%token                        T_CONFLICT_FREE
%token                        T_NOT_CONFLICT_FREE

%token                        T_NO_WAIT
%token                        T_UNCONDITIONAL_WAIT
%token                        T_CONDITIONAL_WAIT
%token                        T_LOOPED_WAIT
%token                        T_WORST_WAIT

%token                        T_SC_FS
%token                        T_SC_PS
%token                        T_SC_NS
%token                        T_SC_US
%token                        T_SC_MS
%token                        T_SC_SEC

%token                        T_EVENT
%token                        T_SC_ZERO_TIME

%token <number_>              T_NUMBER
%token <function_name_>       T_FUNCTION_NAME

%type  <is_conflict_free_>    conflict_annotation
%type  <wait_annotation_>     wait_annotation
%type  <number_>              magnitude
%type  <annotated_time_>      time
%type  <function_annotation_> annotation

%error-verbose

%%

annotation: T_RISC T_FUNCTION_NAME conflict_annotation wait_annotation
            {
              function_annotation = new FunctionAnnotation($2,$3,$4->wait_type_,$4->annotated_time_);
              $$ = function_annotation;
            }
          ;

conflict_annotation: T_CONFLICT_FREE
                     {
                       $$ = true;
                     }
                   | T_NOT_CONFLICT_FREE
                     {
                       $$ = false;
                     }
                   | {
                       $$ = true;
                     }
                   ;

wait_annotation: T_NO_WAIT
                 {
                   WaitAnnotation *wait_annotation = new WaitAnnotation();
                   wait_annotation->wait_type_ = NO_WAIT;
                   wait_annotation->annotated_time_ = NULL;
                   $$ = wait_annotation;
                 }
               | T_UNCONDITIONAL_WAIT time
                 {
                   WaitAnnotation *wait_annotation = new WaitAnnotation();
                   wait_annotation->wait_type_ = UNCONDITIONAL_WAIT;
                   wait_annotation->annotated_time_ = $2;
                   $$ = wait_annotation;
                 }
               | T_CONDITIONAL_WAIT time
                 {
                   WaitAnnotation *wait_annotation = new WaitAnnotation();
                   wait_annotation->wait_type_ = CONDITIONAL_WAIT;
                   wait_annotation->annotated_time_ = $2;
                   $$ = wait_annotation;
                 }
               | T_LOOPED_WAIT time
                 {
                   WaitAnnotation *wait_annotation = new WaitAnnotation();
                   wait_annotation->wait_type_ = LOOPED_WAIT;
                   wait_annotation->annotated_time_ = $2;
                   $$ = wait_annotation;
                 }
               | T_WORST_WAIT time
                 {
                   WaitAnnotation *wait_annotation = new WaitAnnotation();
                   wait_annotation->wait_type_ = WORST_WAIT;
                   wait_annotation->annotated_time_ = $2;
                   $$ = wait_annotation;
                 }
               ;

time: T_TIME T_LPAREN T_NUMBER T_COMMA magnitude T_RPAREN
      {
        AnnotatedTime *annotated_time = new AnnotatedTime();
        annotated_time->units_ = $3;
        annotated_time->magnitude_ = $5;
        $$ = annotated_time;
      }
    | T_EVENT
      {
        AnnotatedTime *annotated_time = new AnnotatedTime();
        annotated_time->event_notification_ = true;
        $$ = annotated_time;
      }
    | T_SC_ZERO_TIME
      {
        AnnotatedTime *annotated_time = new AnnotatedTime();
        annotated_time->sc_zero_time_ = true;
        $$ = annotated_time;
      }
    ;

magnitude: T_SC_FS
           {
             $$ = 1;
           }
         | T_SC_PS
           {
             $$ = 1000;
           }
         | T_SC_NS
           {
             $$ = 1000000;
           }
         | T_SC_US
           {
             $$ = 1000000000;
           }
         | T_SC_MS
           {
             $$ = 1000000000000;
           }
         | T_SC_SEC
           {
             $$ = 1000000000000000;
           }
         ;
%%

extern FILE *function_annotation_parserin;
FunctionAnnotation *function_annotation = NULL;

