#include "class_is_derived_from.h"

#include "rose_nodes.h"
#include "../internal_representation/design.h"
#include "../tools/systemc_definitions.h"

bool
risc::tools::is_derived_from(SgClassDefinition* derived_class,
    SgClassDefinition* base_class)
{
  // FIXME This is a work around for the broken AST link of the base
  // class of sc_event_and_list
  if(base_class == NULL) {
    return false;
  }

  SgBaseClassPtrList base_classes = derived_class->get_inheritances();
  SgBaseClassPtrList::iterator iter = base_classes.begin();

  // iterate all direct base classes
  for(;iter != base_classes.end();iter++) {

    // check current base class
    if((*iter)->get_base_class()->search_for_symbol_from_symbol_table() ==
      base_class->get_declaration()->search_for_symbol_from_symbol_table()) {

      return true;
    }

    SgClassDeclaration *base_class_decl
      = isSgClassDeclaration((*iter)->get_base_class()->get_definingDeclaration());

    if(base_class_decl) {

      // recursive search
      if(is_derived_from(base_class_decl->get_definition(), base_class)) {
        return true;
      }

    } else {

      SgTemplateClassDeclaration *class_template_decl
        = isSgTemplateClassDeclaration((*iter)->get_base_class());

      if(class_template_decl) {

        // In this case we have a class template
        // So, we have to find the instances and check them.

        class_template_decl
          = isSgTemplateClassDeclaration(
              class_template_decl->get_firstNondefiningDeclaration());

        const std::vector<SgNode*> &template_instance_decl
          = RoseNodes::get_template_instance_decl();

        for(std::vector<SgNode*>::const_iterator
            iter  = template_instance_decl.begin();
            iter != template_instance_decl.end();
            iter++) {

          SgTemplateInstantiationDecl *template_class_decl_inst
            = isSgTemplateInstantiationDecl(*iter);

          SgTemplateClassDeclaration *temp
            = template_class_decl_inst->get_templateDeclaration();

          temp = isSgTemplateClassDeclaration(
                   temp->get_firstNondefiningDeclaration());

          if(class_template_decl == temp) {

            template_class_decl_inst
              = isSgTemplateInstantiationDecl(
                  template_class_decl_inst->get_definingDeclaration());

            // recursive search
            if(is_derived_from(template_class_decl_inst->get_definition(),
                  base_class)) {
              return true;
            }
          }
        }

      } else {

        if(!(*iter)->get_base_class()->get_definingDeclaration()) {

          // This can be sometimes quit annoying. So I turned it down.
          //std::cout << "Cannot find any definition of class: "
          //          << SageInterface::get_name((*iter)->get_base_class())
          //          << std::endl;
        }
      }
    }
  }

  return false;
}

bool
risc::tools::is_derived_from_sc_interface(SgClassDefinition* derived_class)
{
  return is_derived_from(derived_class,
             tools::SystemCDefinitions::get_sc_interface());
}

bool
risc::tools::is_derived_from_sc_module(SgClassDefinition* derived_class)
{
  return is_derived_from(derived_class,
             tools::SystemCDefinitions::get_sc_module())
         && !is_derived_from(derived_class, // because of typedef
             tools::SystemCDefinitions::get_sc_channel());
}

bool
risc::tools::is_derived_from_sc_port(SgClassDefinition* derived_class)
{
  return is_derived_from(derived_class,
              tools::SystemCDefinitions::get_sc_port());
}

bool
risc::tools::is_derived_from_sc_port_base(SgClassDefinition* derived_class)
{
  return is_derived_from(derived_class,
             tools::SystemCDefinitions::get_sc_port_base());
}

bool
risc::tools::is_derived_from_sc_port_base(SgVarRefExp* var_ref)
{
  SgClassType *class_type = isSgClassType(var_ref->get_type());

  if(class_type == NULL) {
    return false;
  }

  SgClassDeclaration *class_decl
    = isSgClassDeclaration(class_type->get_declaration());

  class_decl = isSgClassDeclaration(class_decl->get_definingDeclaration());

  SgClassDefinition *class_def = class_decl->get_definition();

  return is_derived_from(class_def,
             tools::SystemCDefinitions::get_sc_port_base());
}

bool
risc::tools::is_derived_from_sc_prim_channel(SgClassDefinition* derived_class)
{
  return is_derived_from(derived_class,
             tools::SystemCDefinitions::get_sc_prim_channel());
}

bool
risc::tools::is_derived_from_sc_channel(SgClassDefinition* derived_class)
{
  return is_derived_from(derived_class,
             tools::SystemCDefinitions::get_sc_channel());
}

bool
risc::tools::is_sc_in_port(SgTemplateInstantiationDefn* template_class)
{
  SgClassDeclaration *class_decl = template_class->get_declaration();
  SgTemplateInstantiationDecl *templ_class_decl =
    isSgTemplateInstantiationDecl(class_decl);

  return templ_class_decl->get_templateDeclaration()->get_type() ==
    tools::SystemCDefinitions::get_sc_in()->get_declaration()->
        get_type();
}

bool
risc::tools::is_sc_out_port(SgTemplateInstantiationDefn* template_class)
{
  SgClassDeclaration *class_decl = template_class->get_declaration();
  SgTemplateInstantiationDecl *templ_class_decl =
    isSgTemplateInstantiationDecl(class_decl);

  return templ_class_decl->get_templateDeclaration()->get_type() ==
    tools::SystemCDefinitions::get_sc_out()->get_declaration()->
        get_type();
}

bool
risc::tools::is_sc_inout_port(SgTemplateInstantiationDefn* template_class)
{
  SgClassDeclaration *class_decl = template_class->get_declaration();
  SgTemplateInstantiationDecl *templ_class_decl =
    isSgTemplateInstantiationDecl(class_decl);

  return templ_class_decl->get_templateDeclaration()->get_type() ==
    tools::SystemCDefinitions::get_sc_inout()->get_declaration()->
        get_type();
}

bool
risc::tools::is_sc_in_clk(SgClassDefinition *definition)
{
  return definition->get_declaration()->get_type() ==
    tools::SystemCDefinitions::get_sc_in_clk()->get_declaration()->
        get_type();
}

bool
risc::tools::is_sc_out_clk(SgClassDefinition *definition)
{
  return definition->get_declaration()->get_type() ==
    tools::SystemCDefinitions::get_sc_out_clk()->get_declaration()->
        get_type();
}

bool
risc::tools::is_sc_inout_clk(SgClassDefinition *definition)
{
  return definition->get_declaration()->get_type() ==
    tools::SystemCDefinitions::get_sc_inout_clk()->get_declaration()->
        get_type();
}

bool
risc::tools::is_sc_event(SgClassDefinition* derived_class)
{
  return is_derived_from(derived_class,
             tools::SystemCDefinitions::get_sc_event())
    || (derived_class->get_declaration()->get_type() ==
         tools::SystemCDefinitions::get_sc_event()->
             get_declaration()->get_type());
}

bool
risc::tools::is_sc_event_and_list(SgClassDefinition* derived_class)
{
  return is_derived_from(derived_class,
             tools::SystemCDefinitions::get_sc_event_and_list())
    || (derived_class->get_declaration()->get_type() ==
         tools::SystemCDefinitions::get_sc_event_and_list()->
             get_declaration()->get_type());
}

bool
risc::tools::is_sc_event_or_list(SgClassDefinition* derived_class)
{
  return is_derived_from(derived_class,
             tools::SystemCDefinitions::get_sc_event_or_list())
    || (derived_class->get_declaration()->get_type() ==
         tools::SystemCDefinitions::get_sc_event_or_list()->
             get_declaration()->get_type());
}

bool
risc::tools::is_sc_event(SgVarRefExp* var_ref)
{
  return var_ref->get_type() ==
    tools::SystemCDefinitions::get_sc_event()->get_declaration()->
        get_type();
}

bool
risc::tools::is_sc_event_and_list(SgVarRefExp* var_ref)
{
  return var_ref->get_type() ==
    tools::SystemCDefinitions::get_sc_event_and_list()
      ->get_declaration()->get_type();
}

bool
risc::tools::is_sc_event_or_list(SgVarRefExp* var_ref)
{
  return var_ref->get_type() ==
    tools::SystemCDefinitions::get_sc_event_or_list()
      ->get_declaration()->get_type();
}

bool
risc::tools::is_sc_in_bool(SgClassDefinition* definition)
{
  SgTemplateInstantiationDefn *temp_inst_def =
    isSgTemplateInstantiationDefn(definition);

  if(temp_inst_def) {

    if(is_sc_in_port(temp_inst_def))  {

      SgTemplateInstantiationDecl *temp_inst_decl =
        isSgTemplateInstantiationDecl(temp_inst_def->get_declaration());

      if(temp_inst_decl->get_templateArguments().size() ==1 &&
          isSgTypeBool((temp_inst_decl->get_templateArguments().at(0))->
              get_type())) {
        return true;
      }
    }
  }

  return false;
}

bool
risc::tools::is_sc_fifo_in(SgVarRefExp* var_ref)
{
  SgClassType *class_type = isSgClassType(var_ref->get_type());

  if(class_type) {

    SgClassDeclaration *class_decl
      = isSgClassDeclaration(class_type->get_declaration());

    class_decl = isSgClassDeclaration(class_decl->get_definingDeclaration());

    SgTemplateInstantiationDecl *template_instance_decl
      = isSgTemplateInstantiationDecl(class_decl);
    if(template_instance_decl) {

      SgTemplateClassDeclaration *template_decl
        = template_instance_decl->get_templateDeclaration();

      template_decl
        = isSgTemplateClassDeclaration(
            template_decl->get_definingDeclaration());

      return tools::SystemCDefinitions::get_sc_fifo_in()
                 == template_decl->get_definition();
    }
  }

  return false;
}

bool
risc::tools::is_sc_fifo_out(SgVarRefExp* var_ref)
{
  SgClassType *class_type = isSgClassType(var_ref->get_type());

  if(class_type) {

    SgClassDeclaration *class_decl
      = isSgClassDeclaration(class_type->get_declaration());

    class_decl = isSgClassDeclaration(class_decl->get_definingDeclaration());

    SgTemplateInstantiationDecl *template_instance_decl
      = isSgTemplateInstantiationDecl(class_decl);
    if(template_instance_decl) {

      SgTemplateClassDeclaration *template_decl
        = template_instance_decl->get_templateDeclaration();

      template_decl
        = isSgTemplateClassDeclaration(
            template_decl->get_definingDeclaration());

      return tools::SystemCDefinitions::get_sc_fifo_out()
                 == template_decl->get_definition();
    }
  }

  return false;
}

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
