#include "systemc.h"

#if 1

typedef struct Particles_t
{
  Particles_t():
    x(0)
  { }

  int x;

  friend std::ostream& operator<<(std::ostream& os, const Particles_t& co);
} particle_t;


//
// user-defined typed double-handshake channel (with transceiver interface)
//
class i_particle_transceiver: virtual public sc_interface
{
public:
  virtual void read(particle_t) = 0;
  virtual void write(const particle_t) = 0;
};

class c_particle_double_handshake_channel:
  public sc_channel, public i_particle_transceiver
{
public:
  c_particle_double_handshake_channel(sc_module_name name):
    sc_channel(name)
  {
    v = false;
    w = false;
  }

  ~c_particle_double_handshake_channel()
  { }

  //#pragma RISC read unconditional-wait event
  void read(particle_t Y)
  {
    if(!v) {
      w = true;
      wait(req);
      w = false;
    }
    Y = tmpd;
    v = false;
    ack.notify(SC_ZERO_TIME);
    wait(ack);
  }

  //#pragma RISC write unconditional-wait event
  void write(const particle_t X)
  {
    tmpd = X;
    v = true;
    if(w) {
      req.notify(SC_ZERO_TIME);
   }
   wait(ack);
  }

private:
  sc_event req;
  sc_event ack;
  bool v, w;
  particle_t tmpd;
};



#if 0
//double handshake implementation for DUTs communication
//
//transceiver interface method
class i_coords_tranceiver: virtual public sc_interface {

public:
  virtual void read(particle_t&) = 0;
  virtual void write(particle_t&) = 0;
};

//Double handshake implementation
class c_coords_double_handshake_channel:
  public sc_channel, // FIXME Tim: was sc_prim_channel before
  public i_coords_tranceiver { //, public i_coords_sender, public i_coords_receiver

  //does it need 2 buffers or 1?

public:
  c_coords_double_handshake_channel(sc_module_name name, int size_ = 4):
      sc_channel(name) {
    size = size_;

    buf = new particle_t[size];
    v = false;
    w = false;
    reset();
  }

  ~c_coords_double_handshake_channel()
  {
    delete [] buf;
  }

  int num_available() const
  {
    return size - free_slots;
  }

  int num_free() const
  {
    return free_slots;
  }

  #pragma RISC write unconditional-wait event
  void write(particle_t& X)
  {
    if(num_free() == 0)
      wait(data_read_event);

    buf[wi] = X;
    wi = (wi+1) % size;
    free_slots --;
    wait(5,SC_NS);
    data_written_event.notify(SC_ZERO_TIME);

    v = true;
    if(w) {
      req.notify(SC_ZERO_TIME);
    }
    wait (ack);
  }

  #pragma RISC read unconditional-wait event
  void read(particle_t& Y)
  {
    if (!v) {
      w = true;
      wait(req);
      w = false;
    }

    /*TODO: check if both waiting for same data lenght*/

    if (num_available() == 0)
        wait(data_written_event);
    Y = buf[ri];
    ri = (ri+1) % size;
    free_slots ++;
    v = false;
    ack.notify(SC_ZERO_TIME);
    wait(ack);
  }

  void reset()
  {
    free_slots = size;
    ri = 0;
    wi = 0;
  }

private:
  int size;       // size
  particle_t *buf;     // circular buffer
  int free_slots; // free space
  int ri;         // read index
  int wi;         // write index

  bool v, w;

  sc_event req;
  sc_event ack;

  sc_event data_read_event;
  sc_event data_written_event;
};

#endif

class Sender: public sc_module {

public:

  SC_HAS_PROCESS(Sender);

  sc_port<i_particle_transceiver> port1;
  sc_port<i_particle_transceiver> port2;

  Sender(sc_module_name mn):
    sc_module(mn)
  {
    SC_THREAD(sender_custom);
  }

  void sender_custom()
  {
    particle_t particle;
    particle.x = 42;
    //std::cout << "writing: " << particle.x << std::endl;
    port1->write(particle);
    //port2->write(particle);

    int c;
    wait(SC_ZERO_TIME);
    int d;
  }
};

class Receiver: public sc_module {

public:

  SC_HAS_PROCESS(Receiver);

  sc_port<i_particle_transceiver> port1;
  sc_port<i_particle_transceiver> port2;

  Receiver(sc_module_name mn):
    sc_module(mn)
  {
    SC_THREAD(receiver_custom);
  }

  void receiver_custom()
  {
    int a;
    wait(SC_ZERO_TIME);
    int b;

    particle_t particle;

    port1->read(particle);
    //port2->read(particle);

    int c;
    wait(SC_ZERO_TIME);
    int d;
    //std::cout << "reading: " << particle.x << std::endl;
  }
};

class Top1: public sc_module {

public:
  Sender sender;
  Receiver receiver;
  c_particle_double_handshake_channel channel1;
  c_particle_double_handshake_channel channel2;

  Top1(sc_module_name mn):
    sc_module(mn),
    sender("sender"),
    receiver("receiver"),
    channel1("channel1"),
    channel2("channel2")
  { }

  void before_end_of_elaboration()
  {
    sender.port1.bind(channel1);
    receiver.port1.bind(channel1);

    sender.port2.bind(channel2);
    receiver.port2.bind(channel2);
  }
};

#endif


///////////////////////////////// fifo example /////////////////////////////////

#if 0

class Sender2: public sc_module {

public:

  SC_HAS_PROCESS(Sender2);

  sc_fifo_out<int> port1;
  sc_fifo_out<int> port2;

  Sender2(sc_module_name mn):
    sc_module(mn)
  {
    SC_THREAD(sender_fifo);
  }

  void sender_fifo()
  {
    int a;
    wait(SC_ZERO_TIME);
    int b;

    //std::cout << "writing: " << 42 << std::endl;
    //std::cout << "writing: " << 43 << std::endl;
    port1.write(42);
    //port2.write(43);

    int c;
    wait(SC_ZERO_TIME);
    int d;
  }
};

class Receiver2: public sc_module {

public:

  SC_HAS_PROCESS(Receiver2);

  sc_fifo_in<int> port1;
  sc_fifo_in<int> port2;

  Receiver2(sc_module_name mn):
    sc_module(mn)
  {
    SC_THREAD(receiver_fifo);
  }

  void receiver_fifo()
  {
    int a;
    wait(SC_ZERO_TIME);
    int b;

    int one = 0;
    int two = 0;

    one = port1.read();
    //two = port2.read();

    //std::cout << "reading: " << one << std::endl;
    //std::cout << "reading: " << two << std::endl;

    int c;
    wait(SC_ZERO_TIME);
    int d;
  }
};

class Top2: public sc_module {

public:
  Sender2 sender;
  Receiver2 receiver;
  sc_fifo<int> channel1;
  sc_fifo<int> channel2;

  Top2(sc_module_name mn):
    sc_module(mn),
    sender("sender"),
    receiver("receiver"),
    channel1("channel1"),
    channel2("channel2")
  { }

  void before_end_of_elaboration()
  {
    sender.port1.bind(channel1);
    receiver.port1.bind(channel1);

    sender.port2.bind(channel2);
    receiver.port2.bind(channel2);
  }
};

#endif

class TopTop: public sc_module {

public:

  Top1 top1;
//  Top1 top2;
//  Top2 top2;

  TopTop(sc_module_name mn):
    sc_module(mn),
    top1("top1")
//    top2("top2")
  { }
};

int sc_main(int argc, char **argv)
{
  TopTop t("top");



  sc_start();
  std::cout << "t.top1.sender.port1: " << (&(t.top1.sender.port1))->get_interface() << std::endl;

  std::cout << "t.top1.sender.port1: " << t.top1.sender.port1.get_interface() << std::endl;
  std::cout << "t.top1.sender.port2: " << t.top1.sender.port2.get_interface() << std::endl;
  std::cout << "t.top1.channel:      " << (sc_interface*)(&t.top1.channel1) << std::endl;


  return 0;
}

