#include <iostream>

#include "risc.h"

int main(int argc, char* argv[])
{
  std::cout << std::endl
            << "start processing file: " __FILE__
            << " ..."
            << std::endl;

  risc::Design d(argc,argv);


  risc::sg::SegmentGraph::FunctionCallBoundarySet function_call_boundary(
      risc::tools::SystemCDefinitions::get_sc_wait_funcs());
  risc::sg::SegmentGraph sg(function_call_boundary);

  for(risc::ModuleVectorIter
      iter_module  = d.module_definitions_.begin();
      iter_module != d.module_definitions_.end();
      iter_module++) {

    std::cout << "Module: " <<  (*iter_module)->get_name() << std::endl;

    for(risc::ThreadVectorIter
        iter_thread = (*iter_module)->threads_.begin();
        iter_thread != (*iter_module)->threads_.end();
        iter_thread++) {

      std::cout << "  " << (*iter_thread)->get_name() << std::endl;


      sg.add_function(*iter_thread);

      //AstDOTGeneration dotgen;
      //dotgen.generate((*iter_thread)->get_ast_node(), (*iter_thread)->get_name());
    }

    std::cout << std::endl;
  }

  sg.clean_graph();
  sg.print_graph("time_advance_table_segment_graph.dot");

  risc::sg::TimeAdvanceTable time_advance_table(sg);

  time_advance_table.print_current_time_advance_table(
                          "current_time_advance_table.html");
  time_advance_table.print_next_time_advance_table(
                          "next_time_advance_table.html");

  std::cout << "end processing file: " __FILE__
            << " ..."
            << std::endl
            << std::endl;

  return 0;
}

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
